/*
 * W2000A Screenshot Utility v1.6.BF.00
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#ifndef WIN32
#include <fcntl.h>
#include <err.h>
#include <termios.h>
#include <time.h>
#else
#include <windows.h>
#include "ComTools.h"
#endif

#define VERSION		00

static void usage(void);
static void set_serial(int);
static int open_serial(char *);
static int rdchar(int);
//static int wrchar(int, char);
static void csv_hdr(FILE *);
static void csv_data(FILE *, unsigned char [4]);
static void csv_ftr(FILE *, int com);
static void ascii_hdr(FILE *);
static void ascii_data(FILE *, unsigned char [4]);
static void ascii_ftr(FILE *, int com);
static void receive_trace(int, int, const char *, void (*)(FILE *), void (*)(FILE *, unsigned char [4]), void (*)(FILE *, int));
static void receive_screen(int);
static void retrieve(int);
static void combine_screen(int);
static time_t gettime(void);


/*
 * XXX: Please note:
 *      This is _not_ 100% identical to the actual
 *      output seen on the DSO.  For example:
 *      The math channel is drawn _over_
 *      the browse bar.  Put the Math channel
 *      behind UI in the planeorder to witness
 *      the actual output.
 */
static const int planeorder[] = {
	7, 8, 9, 10, 		/* Channels 1-4 */
				/* XXX: Red Arrow and Stop-Sign in UI on 10/Channel4? */
	11,			/* Math Channel */
	12, 13,			/* Cursors and Trigger Mark */
	5, 4, 6, 2, 3,		/* UI */
	1,			/* Grid */
	0
};

static const int planecolours[][3] = {
			/* R     G     B */
/* UI_Plane1 */		{ 0xFB, 0xFB, 0xFB }, /* dark white */
/* UI_Plane2 */		{ 0x00, 0x00, 0x00 }, /* black */
/* UI_Plane4 */		{ 0xE6, 0xE6, 0x83 }, /* beige-ish yellow */
/* UI_Plane3 */		{ 0xFF, 0xFD, 0xEE }, /* yellow tinted white */
/* UI_Plane5 */		{ 0x7E, 0x7E, 0x7E }, /* gray */
/* Channel_Plane1 */	{ 0xFF, 0xFF, 0x00 }, /* yellow */
/* Channel_Plane2 */	{ 0x00, 0xFF, 0x00 }, /* light green */
/* Channel_Plane3 */	{ 0x32, 0xAB, 0xFF }, /* light blue */
/* Channel_Plane4 */	{ 0xFF, 0x00, 0x00 }, /* red */
/* Channel_Math_Plane */{ 0xFF, 0x96, 0xF6 }, /* pink */
/* Marker_Plane1 */	{ 0xFF, 0x67, 0x00 }, /* orange-red */
/* Marker_Plane2 */	{ 0xFF, 0x67, 0x00 }, /* orange-red */
/* Grid_Plane */	{ 0x3F, 0x3F, 0x3F }, /* dark gray */
/* Background */	{ 0x00, 0x00, 0x00 }, /* black; don't remove this! */
/* B/W Screenshot */	{ 0xFF, 0xFF, 0xFF }, /* white; don't remove this! */
};
static const unsigned int maxplanes=sizeof(planecolours)/3/sizeof(int);

static char fnamprefix[256];
static unsigned char buffer[640*480/2];					//big enough for color bitmap
static unsigned char *bufptr, *bufend;
static int fnum;
static int rdplanes;
static time_t last_t, now_t;
static int iflag;
static char tracetype[16];

#ifdef WIN32
static void
err(int e, const char *fmt, ...)
{

	/*
	 * Don't even know if win32 has sth. like vsnprintf(),
	 * so we don't care.
	 */
	(void)fprintf(stderr, "Error: %s, exiting.\n", fmt);
	exit(e);
}

static void
errx(int e, const char *fmt, ...)
{

	err(e, fmt);
}
#endif

static void
usage(void)
{

	(void)fprintf(stderr,
			"-----------------------------------------------\n"
			"Usage:\n\n"
#ifdef WIN32
			"scrshot.exe <COM port> \n\noption <COM port> 1 to 9\n\n"

			"Input is read from specified COM port,\n"
			"the default is COM1 if no port is specified.\n\n"
#else
			"./scrshot <Serial device> \n\noption <Serial device> e.g. /dev/ttyS0\n\n"

			"Input is read from serial port, diagnostics and\n"
			"status messages are directed to stderr.\n"
			"Default is /dev/ttyS0 if no device is specified\n\n"
#endif
			"-----------------------------------------------\n\n\n"
	);
}
//---------------------------------------------------------------------------------------------------------------------------------------

static void
set_serial(int com)
{

#ifdef WIN32
	(void)com; /* unused */
#else
	struct termios termios;

	if (tcgetattr(com, &termios) == -1)
		err(1, "tcgetattr() for serial port failed");
	if (cfsetspeed(&termios,B115200) == -1)
		err(1, "cfsetspeed() failed, unable to set desired speed (115200 Baud)");

	cfmakeraw(&termios);
	termios.c_cc[VTIME]=100;
	termios.c_cc[VMIN]=0;
	if (tcsetattr(com, TCSANOW, &termios) == -1)
		err(1, "tcsetattr() failed, unable to set line attributes");
#endif
}
//---------------------------------------------------------------------------------------------------------------------------------------

static int
open_serial(char *device)
{
	int com;

#ifdef WIN32
	if (device == NULL)
		com = 1;
	else
		com = *device - '0';
	if (!ComOpen(com - 1, 115200, P_NONE, S_1BIT, D_8BIT))
		err(1, "Unable to open COM port");
#else
	if (device == NULL)
		device = (char *)"/dev/ttyS0";
	if ((com = open(device, O_RDWR)) == -1)
		err(1, "open(%s) failed", device);
	set_serial(com);
#endif

	return com;
}
//---------------------------------------------------------------------------------------------------------------------------------------
static int
rdchar(int com)
{
#ifdef WIN32
	while (ComGetReadCount(com - 1) < 1)
		Sleep(1);
		//(int)usleep(1000);
	return ComRead(com - 1);
#else
	unsigned char c;

	if (read(com, &c, 1) > 0)
		return c;
	return -1;
#endif
}
//---------------------------------------------------------------------------------------------------------------------------------------
/*
static int
wrchar(int com, char c)
{

#ifdef WIN32
	return ComWrite(com - 1, c);
#else
	return write(com, &c, 1);
#endif
}
*/
//---------------------------------------------------------------------------------------------------------------------------------------
static void
csv_hdr(FILE *fp)
{

	(void)fprintf(fp, "Ch1;Ch2;Ch3;Ch4\n");
}
//---------------------------------------------------------------------------------------------------------------------------------------
static void
csv_data(FILE *fp, unsigned char c[4])
{

	(void)fprintf(fp, "%d;%d;%d;%d\n", c[0], c[1], c[2], c[3]);
}
//---------------------------------------------------------------------------------------------------------------------------------------
static void
csv_ftr(FILE *fp, int com)
{
	unsigned char c[4];
	unsigned char NumberOfChannels;
	int r;

	// get number of channels
	r = rdchar(com);
	NumberOfChannels = (unsigned char)r;

	// get channel status
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "Status\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, ";");	//separator
		
		if (c[i] == 1)
		{ (void)fprintf(fp, "on"); }
		else
		{ (void)fprintf(fp, "off"); }
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, ";n.a.;n.a"); }

	// get voltage ranges
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "\n\nRange\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, ";");	//separator

		switch (c[i])
		{
			case 0: { (void)fprintf(fp, "1mV/div"); break; }
			case 1: { (void)fprintf(fp, "2mV/div"); break; }
			case 2: { (void)fprintf(fp, "5mV/div"); break; }
			case 3: { (void)fprintf(fp, "10mV/div"); break; }
			case 4: { (void)fprintf(fp, "20mV/div"); break; }
			case 5: { (void)fprintf(fp, "50mV/div"); break; }
			case 6: { (void)fprintf(fp, "100mV/div"); break; }
			case 7: { (void)fprintf(fp, "200mV/div"); break; }
			case 8: { (void)fprintf(fp, "500mV/div"); break; }
			case 9: { (void)fprintf(fp, "1V/div"); break; }
			case 10: { (void)fprintf(fp, "2V/div"); break; }
			case 11: { (void)fprintf(fp, "5V/div"); break; }
			case 12: { (void)fprintf(fp, "10V/div"); break; }
			case 13: { (void)fprintf(fp, "20V/div"); break; }
			case 14: { (void)fprintf(fp, "50V/div"); break; }
			case 15: { (void)fprintf(fp, "100V/div"); break; }
		}
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, ";n.a.;n.a"); }

	// get channel coupling
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "\n\nCoupling\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, ";");	//separator
		
		if (c[i] == 2)
		{ (void)fprintf(fp, "DC"); }
		else if (c[i] == 1)
		{ (void)fprintf(fp, "AC"); }
		else
		{ (void)fprintf(fp, "GND"); }
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, ";n.a.;n.a"); }


	// get bandwidth limit
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "\n\nBandwidth Limit\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, ";");	//separator
		
		if (c[i] == 1)
		{ (void)fprintf(fp, "on"); }
		else
		{ (void)fprintf(fp, "off"); }
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, ";n.a.;n.a"); }


	// get channel delay
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "\n\nChannel Delay\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, ";");	//separator
		
		(void)fprintf(fp, "%dns",c[i]);
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, ";n.a.;n.a"); }


	// get ADC setup
	r = rdchar(com); c[0] = (unsigned char)r;

	switch (c[0])
	{
		case 0: (void)fprintf(fp, "\n\nADC-Setup;factory\n"); break;
		case 1: (void)fprintf(fp, "\n\nADC-Setup;high frequ\n"); break;
		case 2: (void)fprintf(fp, "\n\nADC-Setup;test 1\n"); break;
		case 3: (void)fprintf(fp, "\n\nADC-Setup;test 2\n"); break;
		case 4: (void)fprintf(fp, "\n\nADC-Setup;test 3\n"); break;
		case 5: (void)fprintf(fp, "\n\nADC-Setup;test 4\n"); break;
		case 6: (void)fprintf(fp, "\n\nADC-Setup;test 5\n"); break;
	}	

	// get pre gain
	r = rdchar(com); c[0] = (unsigned char)r;

	switch (c[0])
	{
		case 0: (void)fprintf(fp, "\nPre Gain;factory\n"); break;
		case 1: (void)fprintf(fp, "\nPre Gain;gain 1,25\n"); break;
		case 2: (void)fprintf(fp, "\nPre Gain;24 ohm\n"); break;
		case 3: (void)fprintf(fp, "\nPre Gain;33 ohm\n"); break;
		case 4: (void)fprintf(fp, "\nPre Gain;add on\n"); break;
	}	

	// get timebase
	r = rdchar(com); c[0] = (unsigned char)r;

	switch (c[0])
	{
		case 0: //fall throu
		case 1: //fall throu
		case 2: //fall throu
		case 3: //fall throu
		case 4: (void)fprintf(fp, "\nTimebase;1GSa/s\n"); break;
		case 5: (void)fprintf(fp, "\nTimebase;500MSa/s\n"); break;
		case 6: (void)fprintf(fp, "\nTimebase;250MSa/s\n"); break;
		case 9: (void)fprintf(fp, "\nTimebase;25MSa/s\n"); break;
		case 10: (void)fprintf(fp, "\nTimebase;10MSa/s\n"); break;
		case 11: (void)fprintf(fp, "\nTimebase;5MSa/s\n"); break;
		case 12: (void)fprintf(fp, "\nTimebase;2,5MSa/s\n"); break;
		case 13: (void)fprintf(fp, "\nTimebase;1MSa/s\n"); break;
		case 14: (void)fprintf(fp, "\nTimebase;500KSa/s\n"); break;
		case 15: (void)fprintf(fp, "\nTimebase;250KSa/s\n"); break;
		case 16: (void)fprintf(fp, "\nTimebase;100KSa/s\n"); break;
		case 17: (void)fprintf(fp, "\nTimebase;50KSa/s\n"); break;
		case 18: (void)fprintf(fp, "\nTimebase;25KSa/s\n"); break;
		case 19: (void)fprintf(fp, "\nTimebase;10KSa/s\n"); break;
		case 20: (void)fprintf(fp, "\nTimebase;5KSa/s\n"); break;
		case 21: (void)fprintf(fp, "\nTimebase;2,5KSa/s\n"); break;
		case 22: (void)fprintf(fp, "\nTimebase;1KSa/s\n"); break;
		case 23: (void)fprintf(fp, "\nTimebase;500SSa/s\n"); break;
		default: (void)fprintf(fp, "\nTimebase;USTB\n"); break;
	}	

	
}
//---------------------------------------------------------------------------------------------------------------------------------------
static void
ascii_hdr(FILE *fp)
{

	(void)fprintf(fp, "Ch1	Ch2	Ch3	Ch4\n");
}
//---------------------------------------------------------------------------------------------------------------------------------------
static void
ascii_data(FILE *fp, unsigned char c[4])
{

	(void)fprintf(fp, "%d	%d	%d	%d\n", c[0], c[1], c[2], c[3]);
}
//---------------------------------------------------------------------------------------------------------------------------------------
static void
ascii_ftr(FILE *fp, int com)
{
	unsigned char c[4];
	unsigned char NumberOfChannels;
	int r;

	// get number of channels
	r = rdchar(com);
	NumberOfChannels = (unsigned char)r;

	// get channel status
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "Status\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, "	");	//separator
		
		if (c[i] == 1)
		{ (void)fprintf(fp, "on"); }
		else
		{ (void)fprintf(fp, "off"); }
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, "	n.a.	n.a"); }

	// get voltage ranges
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "\n\nRange\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, "	");	//separator

		switch (c[i])
		{
			case 0: { (void)fprintf(fp, "1mV/div"); break; }
			case 1: { (void)fprintf(fp, "2mV/div"); break; }
			case 2: { (void)fprintf(fp, "5mV/div"); break; }
			case 3: { (void)fprintf(fp, "10mV/div"); break; }
			case 4: { (void)fprintf(fp, "20mV/div"); break; }
			case 5: { (void)fprintf(fp, "50mV/div"); break; }
			case 6: { (void)fprintf(fp, "100mV/div"); break; }
			case 7: { (void)fprintf(fp, "200mV/div"); break; }
			case 8: { (void)fprintf(fp, "500mV/div"); break; }
			case 9: { (void)fprintf(fp, "1V/div"); break; }
			case 10: { (void)fprintf(fp, "2V/div"); break; }
			case 11: { (void)fprintf(fp, "5V/div"); break; }
			case 12: { (void)fprintf(fp, "10V/div"); break; }
			case 13: { (void)fprintf(fp, "20V/div"); break; }
			case 14: { (void)fprintf(fp, "50V/div"); break; }
			case 15: { (void)fprintf(fp, "100V/div"); break; }
		}
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, "	n.a.	n.a"); }

	// get channel coupling
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "\n\nCoupling\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, "	");	//separator
		
		if (c[i] == 2)
		{ (void)fprintf(fp, "DC"); }
		else if (c[i] == 1)
		{ (void)fprintf(fp, "AC"); }
		else
		{ (void)fprintf(fp, "GND"); }
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, "	n.a.	n.a"); }


	// get bandwidth limit
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "\n\nBandwidth Limit\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, "	");	//separator
		
		if (c[i] == 1)
		{ (void)fprintf(fp, "on"); }
		else
		{ (void)fprintf(fp, "off"); }
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, "	n.a.	n.a"); }


	// get channel delay
	for (int i=0;i<4;i++){ r = rdchar(com); c[i] = (unsigned char)r; }

	(void)fprintf(fp, "\n\nChannel Delay\n");

	for (int i=0;i<NumberOfChannels;i++) 
	{
		if(i>0) (void)fprintf(fp, "	");	//separator
		
		(void)fprintf(fp, "%dns",c[i]);
	}

	if(NumberOfChannels == 2){ (void)fprintf(fp, "	n.a.	n.a"); }


	// get ADC setup
	r = rdchar(com); c[0] = (unsigned char)r;

	switch (c[0])
	{
		case 0: (void)fprintf(fp, "\n\nADC-Setup	factory\n"); break;
		case 1: (void)fprintf(fp, "\n\nADC-Setup	high frequ\n"); break;
		case 2: (void)fprintf(fp, "\n\nADC-Setup	test 1\n"); break;
		case 3: (void)fprintf(fp, "\n\nADC-Setup	test 2\n"); break;
		case 4: (void)fprintf(fp, "\n\nADC-Setup	test 3\n"); break;
		case 5: (void)fprintf(fp, "\n\nADC-Setup	test 4\n"); break;
		case 6: (void)fprintf(fp, "\n\nADC-Setup	test 5\n"); break;
	}	

	// get pre gain
	r = rdchar(com); c[0] = (unsigned char)r;

	switch (c[0])
	{
		case 0: (void)fprintf(fp, "\nPre Gain	factory\n"); break;
		case 1: (void)fprintf(fp, "\nPre Gain	gain 1,25\n"); break;
		case 2: (void)fprintf(fp, "\nPre Gain	24 ohm\n"); break;
		case 3: (void)fprintf(fp, "\nPre Gain	33 ohm\n"); break;
		case 4: (void)fprintf(fp, "\nPre Gain	add on\n"); break;
	}	

	// get timebase
	r = rdchar(com); c[0] = (unsigned char)r;

	switch (c[0])
	{
		case 0: //fall throu
		case 1: //fall throu
		case 2: //fall throu
		case 3: //fall throu
		case 4: (void)fprintf(fp, "\nTimebase	1GSa/s\n"); break;
		case 5: (void)fprintf(fp, "\nTimebase	500MSa/s\n"); break;
		case 6: (void)fprintf(fp, "\nTimebase	250MSa/s\n"); break;
		case 9: (void)fprintf(fp, "\nTimebase	25MSa/s\n"); break;
		case 10: (void)fprintf(fp, "\nTimebase	10MSa/s\n"); break;
		case 11: (void)fprintf(fp, "\nTimebase	5MSa/s\n"); break;
		case 12: (void)fprintf(fp, "\nTimebase	2,5MSa/s\n"); break;
		case 13: (void)fprintf(fp, "\nTimebase	1MSa/s\n"); break;
		case 14: (void)fprintf(fp, "\nTimebase	500KSa/s\n"); break;
		case 15: (void)fprintf(fp, "\nTimebase	250KSa/s\n"); break;
		case 16: (void)fprintf(fp, "\nTimebase	100KSa/s\n"); break;
		case 17: (void)fprintf(fp, "\nTimebase	50KSa/s\n"); break;
		case 18: (void)fprintf(fp, "\nTimebase	25KSa/s\n"); break;
		case 19: (void)fprintf(fp, "\nTimebase	10KSa/s\n"); break;
		case 20: (void)fprintf(fp, "\nTimebase	5KSa/s\n"); break;
		case 21: (void)fprintf(fp, "\nTimebase	2,5KSa/s\n"); break;
		case 22: (void)fprintf(fp, "\nTimebase	1KSa/s\n"); break;
		case 23: (void)fprintf(fp, "\nTimebase	500SSa/s\n"); break;
		default: (void)fprintf(fp, "\nTimebase	USTB\n"); break;
	}	

	
}
//---------------------------------------------------------------------------------------------------------------------------------------

static void
receive_trace(int com, int type, const char *ftype, void fun_hdr(FILE *), void fun_data(FILE *, unsigned char [4]), void fun_ftr(FILE *, int))
{
	FILE *fp;
	unsigned char c[4];
	char fnam[256];
	int r, chan, idx;

	(void)type; /* unused */

	(void)snprintf(fnam, sizeof(fnam), "%s-%.4d.%s", fnamprefix, fnum, ftype);
	(void)fprintf(stderr, "  - Receiving trace data (file: %s)\n", fnam);
	if ((fp = fopen(fnam, "w+")) == NULL)
		err(1, "fopen(%s, w+) failed", fnam);
	fnum++;

	//header line
	fun_hdr(fp);	

	//ADC-data of all channels
	for (idx=0;idx<16384;idx++) {
		for (chan=0;chan<4;chan++) {
			r = rdchar(com);
			c[chan] = (unsigned char)r;
		}
		fun_data(fp, c);
	}

	//empty line
	(void)fprintf(fp, "\n");

	//actual DSO-Parameters on footer
	fun_ftr(fp, com);

	if (fclose(fp) != 0)
		err(1, "fclose(%s) failed", fnam);
}
//---------------------------------------------------------------------------------------------------------------------------------------
static void
receive_screen(int com)
{
	
	unsigned int llength;			/* RLE */
	unsigned char lastbyte, esc;
	int outcnt, incnt;
	int r;
	unsigned char c;
	unsigned int i;
	unsigned int bufsize;

	bufsize=640*480/2;
	rdplanes=maxplanes;
	(void)fprintf(stderr, "  - Receiving a colour screenshot\n");

	/*********************************************************************
	* Read all from stdin/COMx, decompress and fill buffer
	*********************************************************************/
	incnt = outcnt = 0;
	llength = 0;
	lastbyte = esc = 0;
	bufptr= buffer;
	while (((r = rdchar(com)) >= 0) && (bufptr<buffer+bufsize)) {
		c=(unsigned char)r;
		incnt++;
		if (incnt % 100 == 0)
			(void)fprintf(stderr, "\r    %d bytes...", incnt);

		if (esc==2) {
			llength|=(c);					//2nb Byte after ESC is MSB
			for (i=0;i<llength;i++) {
				*bufptr++=lastbyte;
				outcnt++;
			}
			llength=0;
			esc=0; }
		else if (esc==1) {					//2nd escape
			if (c==0xFF) {					//escaped 0xFF
				*bufptr++=c;
				outcnt++;
				lastbyte=c;
				llength=0;
				esc=0;					//Escape-Sequence finished
			} 
			else if (c==0xFE) {		//EOF
				(void)fprintf(stderr, "\nEOF\n");
				bufptr--;						// remove 0x00, not necessary
				break; }
			else {					//1st Byte after ESC is MSB
				llength=c<<8;
				esc++; }
			}
		else if (c != 0xFF) {
			*bufptr++=c;
			outcnt++;
			lastbyte=c;
		} else
			esc=1;
	}
	(void)fprintf(stderr, "\n");
	bufend=bufptr;
	bufptr=buffer;
	
	(void)fprintf(stderr, "* Total bytes transferred: %d(%d) ", outcnt, incnt);
	now_t = gettime();
	(void)fprintf(stderr, "in %lds\n", now_t-last_t);
	last_t = now_t;

	

}
//-------------------------------------------------------------------------------------------------------------------------------------------
static void
retrieve(int com)
{
	int dumptype, r;
	unsigned char c, l;
	int pgm = 0;	//PGM

	(void)fprintf(stderr, "* Waiting for Screenshot or Trace...\n");

	last_t = gettime();
	l = 0;

	// wait for start marker from DSO
	for(;;)
 	{
		r = rdchar(com);
		c = (unsigned char)r;
		if (c == 0xFF) 	//second start marker
		{
			if (l == 'B' || l == 'S') break;// start of BMP transmission
			else if (l == 'P') break;	// start of PGM transmission
			else if (l == 'C') break;	// start of CSV transmission
			else if (l == 'A') break;	// start of ASCII transmission
		}
		l = c;
	}

	//if (r <= 0) exit(1);

	(void)fprintf(stderr, "* Received Start Marker");
	now_t = gettime();
	(void)fprintf(stderr, " after %lds\n", now_t-last_t);
	last_t = now_t;

	/*********************************************************************
	* Next char determines type of data
	*********************************************************************/
	dumptype = (int)rdchar(com);

	switch (dumptype)
	{
		case 0: //screenshot  => FALLTHRU to 1
		case 1:	
			if (l == 'B' || l == 'S') pgm = 0;	//BMP -> S is OS-compatibility
			else if (l == 'P') pgm = 1;		//PGM
	
			receive_screen(com);
			combine_screen(!pgm);
			break;
		case 2: //CSV or ASCII dump
			if (l=='C' || 'S')		//CSV
			{ receive_trace(com, dumptype, "csv", csv_hdr, csv_data, csv_ftr); }
			else if (l=='A')		//ASCII
			{ receive_trace(com, dumptype, "asc", ascii_hdr, ascii_data, ascii_ftr); }
			else
			{ errx(1, "Internal Error:  Unhandled Trace Filetype in retrieve() (at 2): %s", tracetype); }
			break;
		
		case 3:	//ASCII dump -> OS-compatibility
			if (l=='S' || l=='A')
			{ receive_trace(com, dumptype, "asc", ascii_hdr, ascii_data, ascii_ftr); }
			else
			{ errx(1, "Internal Error:  Unhandled Trace Filetype in retrieve() (at 2): %s", tracetype); }
			break;


		default:
			errx(1, "Internal Error:  Unhandled Dumptype in retrieve() (at 2): %d", dumptype);
	
	}
}
//--------------------------------------------------------------------------------------------------------------------------------------------
static void
combine_screen(int bmp)
{
	#define INVP(colourcomp)	iflag ? (255 - (colourcomp)) : (colourcomp)

	FILE *fp;
	char fnam[256], suf[4];
	unsigned int colour;
	unsigned int x;
	
	
	if (bmp)
	(void)snprintf(suf, sizeof(suf), "bmp");
	else
	(void)snprintf(suf, sizeof(suf), "ppm");


	(void)snprintf(fnam, sizeof(fnam), "%s-%.4d.%s", fnamprefix, fnum, suf);
	(void)fprintf(stderr, "* Combining (file: %s)...\n", fnam);
	if ((fp = fopen(fnam, "w+")) == NULL)
		err(1, "fopen(%s, w+) failed", fnam);
	fnum++;

	bufptr=buffer;

	if (bmp)
	{
		/*
		 * http://en.wikipedia.org/wiki/BMP_file_format
		 */
		/* 0x00 */ (void)fprintf(fp, "BM");
		/* 0x02 */ (void)fprintf(fp, "%c%c%c%c%c%c%c%c", 0, 0, 0, 0, 0, 0, 0, 0);
		/* 0x0A */ (void)fprintf(fp, "%c%c%c%c", 54, 0, 0, 0);		/* 40 (header size) + 0x0A + 4 */
		/* 0x0E */ (void)fprintf(fp, "%c%c%c%c", 40, 0, 0, 0);		/* header size */
		/* 0x12 */ (void)fprintf(fp, "%c%c%c%c", 128, 2, 0, 0);		/* 640px */
		/* 0x16 */ (void)fprintf(fp, "%c%c%c%c", 32, 254, 255, 255);	/* -480px */		//may not work on linux
		/* 0x1A */ (void)fprintf(fp, "%c%c", 1, 0);
		/* 0x1C */ (void)fprintf(fp, "%c%c", 24, 0);			/* bits per pixel */
		/* 0x1E */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);		/* no compression */
		/* 0x22 */ (void)fprintf(fp, "%c%c%c%c", 0, 16, 14, 0);		/* bitmap data size: 640*480*3=921600 Bytes */	//
		/* 0x26 */ (void)fprintf(fp, "%c%c%c%c", 19, 11, 0, 0);	//KB	//0,0,0,0
		/* 0x2A */ (void)fprintf(fp, "%c%c%c%c", 19, 11, 0, 0);	//KB	//0,0,0,0
		/* 0x2E */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);
		/* 0x32 */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);

		for(x=0;x<640*480/2;x++)
		{
			colour=bufptr[x] & 0x0F;
			/* colour order BGR */
			(void)fputc(INVP(planecolours[colour][2]), fp);
			(void)fputc(INVP(planecolours[colour][1]), fp);
			(void)fputc(INVP(planecolours[colour][0]), fp);

			colour=bufptr[x] >> 4;
			/* colour order BGR */
			(void)fputc(INVP(planecolours[colour][2]), fp);
			(void)fputc(INVP(planecolours[colour][1]), fp);
			(void)fputc(INVP(planecolours[colour][0]), fp);
		}
		
	}
	else	//PPM -> Portable Pixel Map
	{
		/*
		 * http://netpbm.sourceforge.net/doc/ppm.html
		 */
		(void)fprintf(fp, "P6\n640 480\n255 ");
		
		for(x=0;x<640*480/2;x++)
		{
			colour=bufptr[x] & 0x0F;
			/* colour order RGB */
			(void)fputc(INVP(planecolours[colour][0]), fp);
			(void)fputc(INVP(planecolours[colour][1]), fp);
			(void)fputc(INVP(planecolours[colour][2]), fp);

			colour=bufptr[x] >> 4;
			/* colour order RGB */
			(void)fputc(INVP(planecolours[colour][0]), fp);
			(void)fputc(INVP(planecolours[colour][1]), fp);
			(void)fputc(INVP(planecolours[colour][2]), fp);
		}
	}

	if (fclose(fp) != 0)
	{ err(1, "fclose(%s) failed", fnam); }




#ifdef WIN32
//	puts("\a");
	printf ("\a") ;
//	Beep(523,500);
	Sleep(500);
// 	usleep(500);
//	Beep(523,500);
	printf ("\a") ;
//	puts("\a");
#endif


	(void)fprintf(stderr, "* Done\n");
	(void)fprintf(stderr, "********************************************\n");
	(void)fprintf(stderr, "\n");


}
//------------------------------------------------------------------------------------------------------------------------------
static time_t
gettime(void)
{
#ifdef WIN32
	/*
	 * Yes, this is crude.  But I won't implement
	 * a decent API just for Windows.
	 */
	SYSTEMTIME st;

	GetSystemTime(&st);
	return st.wDay * 24*3600 + st.wHour * 3600 + st.wMinute * 60 + st.wSecond;
#else
	return time(NULL);
#endif
}
//------------------------------------------------------------------------------------------------------------------------------
int
main(int argc, char **argv)
{
	char *device;
	int com;

	device = NULL;
	fnum = 0;
	(void)snprintf(tracetype, sizeof(tracetype), "csv");
	(void)snprintf(fnamprefix, sizeof(fnamprefix), "screenshot");

	//If no arguments are found, print help
	if(argc<2)
 	{ usage(); }


	// get option argument
	device = argv[1];

	(void)fprintf(stderr, "********************************************\n");
	(void)fprintf(stderr, "*          Screenshot V2.0.BF.%d            *\n",VERSION);
	(void)fprintf(stderr, "********************************************\n");
	
	com = open_serial(device);
	set_serial(com);

	if (device==NULL)
	(void)fprintf(stderr, "Listening on default port\n");
	else
	(void)fprintf(stderr, "Listening on %s\n",device);

	for(;;)
	{ retrieve(com); }		//get data from DSO
	
	return 0;
}

 	  	 
